unit iHiLite;

interface

uses
    SysUtils, Classes, Contnrs, Graphics;

type
    //--------------------------------------------------------------------------
    //  \NX
    //--------------------------------------------------------------------------
    THighlight = class
    private
        FText: string;
        FFoundPos: integer;
    protected
        FColor: TColor;
        FBold: Boolean;
        function    GetLength(): integer; virtual; abstract;
        function    Match(const S: string; const Offset: integer;
                          var FoundPos: integer): Boolean; virtual; abstract;
    public
        constructor Create(const S: string; Color: integer; Bold: Boolean); virtual;
        property Len: integer read GetLength;
        property FoundPos: integer read FFoundPos;
    end;

    THighlightClass = class of THighlight;

    //--------------------------------------------------------------------------
    //  萔NX
    //--------------------------------------------------------------------------
    THighlightString = class(THighlight)
    private
        FText: char;
        FLen: integer;
    protected
        function    GetLength(): integer; override;
        function    Match(const S: string; const Offset: integer;
                          var FoundPos: integer): Boolean; override;
    public
        constructor Create(const S: string; Color: integer; Bold: Boolean); override;
    end;

    //--------------------------------------------------------------------------
    //  1sRgNX
    //--------------------------------------------------------------------------
    THighlightComment = class(THighlight)
    private
        FLen: integer;
    protected
        function    GetLength(): integer; override;
        function    Match(const S: string; const Offset: integer;
                          var FoundPos: integer): Boolean; override;
    end;

    //--------------------------------------------------------------------------
    //  \ꏈNX
    //--------------------------------------------------------------------------
    THighlightReserved = class(THighlight)
    protected
        function    GetLength(): integer; override;
        function    Match(const S: string; const Offset: integer;
                          var FoundPos: integer): Boolean; override;
    end;

    //--------------------------------------------------------------------------
    //  e
    //--------------------------------------------------------------------------
    THighlightManager = class
    private
        FWorkList: TStringList;
        FHighlightList: TObjectList;
        procedure   LoadHighlightList();
    public
        constructor Create();
        destructor  Destroy(); override;
        procedure   DrawLine(const Line: string; Canvas: TCanvas;
                             const PosX, PosY, CharWidth: integer);
    end;

var
    HiLiteMan: THighlightManager = nil;

implementation

//-----------------------------------------------------------------------------
//  "<"">""&"AHTMLŐ\邽߂̕ϊ
function ChangeHTMLSpecialChar(text: string): string;
var
    i: integer;
    src, dest: array[0..2] of string;
begin

    //ȂƂȂ[vł܂킵Ȃ̂ŔzɁB
    //"&"ŏɕϊƂȂƑςȂƂ
    src[0] := '&';  dest[0] := '&amp;';
    src[1] := '<';  dest[1] := '&lt;';
    src[2] := '>';  dest[2] := '&gt;';

    //ϊRs[
    Result := Text;
    //bAfor i := 0 to 2 do
    for i := Low(src) to High(src) do
    begin
        //uĂ邾
        Result := StringReplace(Result, src[i], dest[i], [rfReplaceAll]);
    end;

end;

//-----------------------------------------------------------------------------
//  AvP[Vf[^tH_
function    GetDataDir(): string;
begin
    Result := ExtractFilePath(ParamStr(0)) + 'dat\';
end;

//*****************************************************************************
//  THighlight

//-----------------------------------------------------------------------------
//  RXgN^
constructor THighlight.Create(const S: string; Color: integer; Bold: Boolean);
begin
    FText := S;
    FColor := Color;
    FBold := Bold;
end;

//*****************************************************************************
//  THighlightString

//-----------------------------------------------------------------------------
//  RXgN^
constructor THighlightString.Create(const S: string; Color: integer; Bold: Boolean);
begin
    FText := S[1];
    FColor := Color;
    FBold := Bold;
end;

//-----------------------------------------------------------------------------
//  ̒
function    THighlightString.GetLength(): integer;
begin
    Result := FLen;
end;

//-----------------------------------------------------------------------------
//  S̕񌟍
function    THighlightString.Match(const S: string;   //  ׂ1s
                  const Offset: integer;        //  Jnʒu
                  var FoundPos: integer         //  ʒu
                  ): Boolean;                   //  ?
var
    Target: string;
    i, Index, RealIndex, LastPos: integer;
    c, nextc: char;
begin
    Result := false;

    //  ۂ̌Ώە
    Target := Copy(S, Offset, Length(S));

    //  Ƃ肠KɒT
    Index := AnsiPos(UpperCase(FText), UpperCase(Target));
    //  łȂI
    if (Index > 0) then
    begin
        //  ۂ̈ʒu
        RealIndex := Offset + Index - 1;
        //  Pcʒu
        LastPos := RealIndex; 
        //  Ă܂傤
        i := RealIndex + 1;
        while (i <= Length(S)) do
        begin
            //  1oCgłȂƑʖ
            if (ByteType(S, i) <> mbSingleByte) then
            begin
                Inc(i);
                continue;
            end;

            c := S[i];

            //  ̏I肾낤?
            if (c <> FText) then
            begin
                Inc(i);
                continue;
            end;

            //  2AƂĂȂ낤˂
            if (i < Length(S)) then
            begin
                nextc := S[i + 1];
                if (nextc = FText) then
                begin
                    Inc(i, 2);
                    continue;
                end;
            end;

            //  ܂ł݂
            LastPos := i;
            break;
        end;

        //  ܂ȂƂ?
        if (RealIndex = LastPos) then LastPos := Length(S);

        //  ǂ
        FLen := LastPos - RealIndex + 1;

        //@̑
        FoundPos := RealIndex;
        FFoundPos := FoundPos;
        Result := true;
    end;
end;

//*****************************************************************************
//  THighlightComment

//-----------------------------------------------------------------------------
//  ̒
function    THighlightComment.GetLength(): integer;
begin
    Result := FLen;
end;

//-----------------------------------------------------------------------------
//  S̕񌟍
function    THighlightComment.Match(const S: string;   //  ׂ1s
                  const Offset: integer;        //  Jnʒu
                  var FoundPos: integer         //  ʒu
                  ): Boolean;                   //  ?
var
    Target: string;
    Index, RealIndex, LastPos: integer;
begin
    Result := false;

    //  ۂ̌Ώە
    Target := Copy(S, Offset, Length(S));

    //  T
    Index := AnsiPos(UpperCase(FText), UpperCase(Target));
    //  łȂI
    if (Index > 0) then
    begin
        //  ۂ̈ʒu
        RealIndex := Offset + Index - 1;
        //  Ō܂łׂđΏ
        LastPos := Length(S); 
        //  ǂ
        FLen := LastPos - RealIndex + 1;
        //@̑
        FoundPos := RealIndex;
        FFoundPos := FoundPos;
        Result := true;
    end;
end;

//*****************************************************************************
//  THighlightReserved

//-----------------------------------------------------------------------------
//  ̒
function    THighlightReserved.GetLength(): integer;
begin
    Result := Length(FText);
end;

//-----------------------------------------------------------------------------
//  S̕񌟍
function    THighlightReserved.Match(const S: string;   //  ׂ1s
                  const Offset: integer;        //  Jnʒu
                  var FoundPos: integer         //  ʒu
                  ): Boolean;                   //  ?
var
    Target: string;
    Index, RealIndex: integer;
begin
    Result := false;
    //  ۂ̌Ώە
    Target := Copy(S, Offset, Length(S));

    while (Target <> '') do
    begin
        //  Ƃ肠KɒT
        Index := AnsiPos(UpperCase(FText), UpperCase(Target));
        //  łȂI
        if (Index = 0) then
        begin
            break;
        end
        //  ܂
        else
        begin
            //  ۂ̈ʒu
            RealIndex := Offset + Index - 1;
            //  O̕ɕςȂ̂˂?
            if (RealIndex > 1) then
            begin
                //  ȊO̕ʖڂł
                if not (S[RealIndex - 1] in [' ',#9,'(',')',';']) then
                begin
                    //  ͈͂ɂ炷
                    Target := Copy(Target, Index + Len, Length(S));
                    continue;
                end;
            end;
            //  ̕ɕςȂ̂˂?
            if (RealIndex + Len - 1 < Length(S)) then
            begin
                //  ȊO̕ʖڂł
                if not (S[RealIndex + Len] in [' ', #9, '(', ')',';']) then
                begin
                    //  ͈͂ɂ炷
                    Target := Copy(Target, Index + Len, Length(S));
                    continue;
                end;
            end;
            //  ̗p
            FoundPos := RealIndex;
            FFoundPos := FoundPos;
            Result := true;
            break;
        end;
    end;
end;

//*****************************************************************************
//  TSqlToHtml

//-----------------------------------------------------------------------------
//  RXgN^
constructor THighlightManager.Create();
begin
    //  Ɨp
    FWorkList := TStringList.Create();
    //  \ꏈNXXg
    FHighlightList := TObjectList.Create(true);
    //  \NX쐬
    LoadHighlightList();
end;
//-----------------------------------------------------------------------------
//  fXgN^
destructor  THighlightManager.Destroy();
begin
    FHighlightList.Free();
    FWorkList.Free();
end;
//-----------------------------------------------------------------------------
//  \ꗗǍ
procedure   THighlightManager.LoadHighlightList();

    //  ------------------------------
    function GetHighLightClass(const Name: string): THighlightClass;
    begin
        if (Name = 'LINECOMMENT')  then Result := THighlightComment      else
        if (Name = 'STRING')       then Result := THighlightString       else
        if (Name = 'WORD')         then Result := THighlightReserved     else
        Result := nil;
    end;

    //  ------------------------------
    procedure CreateHighlights(HighlightClass: THighlightClass; const Path: string);
    var
        F: TextFile;
        S: string;
        Color: integer;
        Bold: Boolean;
    begin
        AssignFile(F, Path);
        try
            //  J
            Reset(F);
            //  I肶Ȃ
            if (Eof(F)) then exit;

            //  1sځF
            Readln(F, S);
            Color := StrToIntDef(S, 0);
            //  I肶Ȃ
            if (Eof(F)) then exit;

            //  2sځ{[hǂ
            Readln(F, S);
            Bold := StrToBoolDef(S, false);

            //  cǍ
            while not Eof(F) do
            begin
                //  ǂ
                Readln(F, S);
                //  ȒPɃ`FbN
                if (S = '') then continue;
                if (HighlightClass = THighlightString) then
                    if (Length(S) <> 1) then continue;
                //  NX쐬Ă
                FHighlightList.Add(HighlightClass.Create(S, Color, Bold));
            end;
        finally
            CloseFile(F);
        end;
    end;

var
    Entry: TStringList;
    Helper: TStringList;
    HighlightClass: THighlightClass;
    EntryFile: string;

    i: integer;
begin

    //  ------------------------------
    //  Ô߃NA
    FHighlightList.Clear();

    //  ------------------------------
    //  ƗpStringList쐬
    Entry := TStringList.Create();
    Helper := TStringList.Create();

    //  ------------------------------
    //  Gg[t@CǍ
    Entry.LoadFromFile(GetDataDir() + '__entry.txt');

    //  ------------------------------
    //  es
    for i := 0 to Entry.Count - 1 do
    begin
        //  1s𕪉
        Helper.CommaText := Entry[i];

        //  \NXʃeLXgEt@C2łȂΔ΂
        if (Helper.Count <> 2) then continue;

        //  ΉNXȂĂ΂
        HighlightClass := GetHighLightClass(Helper[0]);
        if (HighlightClass = nil) then continue;

        //  t@C݂ĂȂĂ΂
        EntryFile := GetDataDir() + Helper[1];
        if (not FileExists(EntryFile)) then continue;

        //  Ă
        CreateHighlights(HighlightClass, EntryFile);
    end;

    //  ------------------------------
    //  
    Helper.Free;
    Entry.Free;
end;
//-----------------------------------------------------------------------------
//  1sϊ
procedure   THighlightManager.DrawLine(const Line: string; Canvas: TCanvas;
                     const PosX, PosY, CharWidth: integer);

    procedure _DrawText(const s: string; var aPosX, aPosY: integer);
    var
        i: integer;
        MbType: TMbcsByteType;
    begin
        //  ` - 1
        for i := 1 to Length(s) do
        begin
            //  oCg^Cv́H
            MbType := ByteType(s, i);
            //  1oCgȂ炻̂܂
            if (MbType = mbSingleByte) then
            begin
                Canvas.TextOut(aPosX, aPosY, s[i]);
                Inc(aPosX, CharWidth);
            end
            //  2oCg2oCgڂȂ
            else if (MbType = mbTrailByte) then
            begin
                //  1O̕ƍ킹
                Canvas.TextOut(aPosX, aPosY, s[i - 1] + s[i]);
                Inc(aPosX, 2 * CharWidth);
            end;
        end;
    end;

var
    i: integer;
    H, OkH: THighlight;
    Offset, FoundPos: integer;
    NormalText, HiLiteText: string;
    aPosX, aPosY: integer;

begin
    aPosX := PosX;
    aPosY := PosY;

    //  Jnʒu
    Offset := 1;
    //  ̗pꂽ\
    OkH := nil;

    //  Jnʒu̒𒴂瓖RI
    while (Offset <= Length(Line)) do
    begin

        //  e\Ɍ
        for i := 0 to FHighlightList.Count - 1 do
        begin
            H := THighlight(FHighlightList[i]);
            //  
            if (H.Match(Line, Offset, FoundPos)) then
            begin
                //  O߂?
                if (OkH = nil) then
                begin
                    //  oĂ
                    OkH := H;
                end
                //  ߂ĂȂ
                else
                begin
                    //  O̓z葁ʒuŌ?
                    if (FoundPos < OkH.FoundPos) then
                    begin
                        //  O̗p
                        OkH := H;
                    end;
                end;
            end;
        end;
        //  ̂
        if (OkH <> nil) then
        begin
            //  ʏ\(Jnʒu甭ʒu܂)
            NormalText := Copy(Line, Offset, OkH.FoundPos - Offset);
            //  \Ώ(ʒu甭)
            HiLiteText := Copy(Line, OkH.FoundPos, OkH.Len);

            //  ʏ`
            Canvas.Font.Color := clBlack;
            Canvas.Font.Style := [];
            _DrawText(NormalText, aPosX, aPosY);

            //  `
            Canvas.Font.Color := OkH.FColor;
            if OkH.FBold then Canvas.Font.Style := [fsBold];
            _DrawText(HiLiteText, aPosX, aPosY);

            //  ɖ߂
            Canvas.Font.Color := clBlack;
            Canvas.Font.Style := [];

            //  ̗\͔\ŒT
            Offset := OkH.FoundPos + OkH.Len;
            OkH := nil;
        end
        //  Ȃ
        else
        begin
            //  ʏ\(ʒuŌ܂)
            NormalText := Copy(Line, Offset, Length(Line));

            //  ʏ`
            Canvas.Font.Color := clBlack;
            Canvas.Font.Style := [];
            _DrawText(NormalText, aPosX, aPosY);

            //  ̍s͒אs
            break;
        end;
    end;
end;

//-----------------------------------------------------------------------------
initialization
    HiLiteMan := THighlightManager.Create();
    HiLiteMan.LoadHighlightList();
//-----------------------------------------------------------------------------
finalization
    HiLiteMan.Free();
//-----------------------------------------------------------------------------
end.

